(* Programming and Reasoning with Infinite Data in Isabelle/HOL (POPL 2018) *)
(* Mathias Fleury, Andreas Lochbihler, Andrei Popescu *)

theory Codatatypes imports Prelim
begin
  
codatatype 'a stream = SCons (shd: 'a) (stl: "'a stream")
  
(* Many useful operators are defined, e.g., *)
  
term map_stream
term rel_stream
term set_stream
  
(* ... and many general-purpose theorems about them are proved:  *)
      
(* Compositionality, e.g.: *)
thm stream.map_comp stream.rel_compp
  
(* Monotonicity: *)
thm stream.rel_mono

(* Connections between the operators, e.g.: *)
thm stream.map_cong  
  
(* Relational parametricity of all the operators: *)
find_theorems name: stream name: transfer
  
  
(****************************************************)
  
(* All this infrastructure (operators and theorems) are 
maintained when nesting (co)datatypes:  *)

bnf_axiomatization 'a B 
datatype 'a tree = Leaf 'a | Branch "'a tree B" 
  
(* E.g., just like streams, trees are functoprs and relators: *)
term map_tree term rel_tree
 
thm tree.map_comp tree.rel_compp

  
(* How do we induct over trees? Using the stream set operator! *)
thm tree.induct
  

    
(* And so on... can nest everything in everything *)
codatatype 'a crazy = Node "'a crazy tree"
  
(* Can also nest non-codatatypes, if they have the 
right operators on them. E.g.: *)
  
codatatype 'a crazier = Node "'a crazier tree fset multiset"
  


  
(* Bottom line: Isabelle/HOL works hard to "know" its (co)datatypes  *)

  
end