(* Programming and Reasoning with Infinite Data in Isabelle/HOL (POPL 2018) *)
(* Mathias Fleury, Andreas Lochbihler, Andrei Popescu *)

theory Coinduction imports Prelim
begin
  
codatatype 'a stream = SCons (shd: 'a) (stl: "'a stream")

lemma 
fixes xs1 xs2 :: "nat stream"
assumes "R xs1 xs2" shows "xs1 = xs2"
using assms proof (coinduction arbitrary: xs1 xs2)
  case Eq_stream
  thus ?case apply simp sorry
qed 

(* Above, the following is invoked: *)
thm stream.coinduct

  
corec (friend) plus :: "nat stream \<Rightarrow> nat stream \<Rightarrow> nat stream" where 
  "plus xs ys = SCons (shd xs + shd ys) (plus (stl xs) (stl ys))"

lemma plus_dest[simp]:
"shd (plus xs ys) = shd xs + shd ys"
"stl (plus xs ys) = plus (stl xs) (stl ys)"
by (subst plus.code, simp)+ 

(* Concrete example proof: *)
lemma plus_comm:  
"plus xs1 xs2 = plus xs2 xs1"
apply (coinduction arbitrary: xs1 xs2) 
by auto

corec (friend) times :: "nat stream \<Rightarrow> nat stream \<Rightarrow> nat stream" where 
  "times xs ys = SCons (shd xs * shd ys) (plus (times xs (stl ys)) (times (stl xs) ys))"

lemma times_dest[simp]: 
"shd (times xs ys) = shd xs * shd ys"
"stl (times xs ys) = plus (times xs (stl ys)) (times (stl xs) ys)"
by(subst times.code, simp)+

lemma 
fixes xs1 xs2 :: "nat stream"
assumes "R xs1 xs2" shows "xs1 = xs2"
using assms proof (coinduction arbitrary: xs1 xs2 rule: stream.coinduct_upto)
  case Eq_nat_stream  
  thus ?case apply simp  sorry
qed

(* Above, the following is invoked: *) 
thm stream.coinduct_upto(1)

term stream.v2.congclp
(* includes the original predicate: *)
thm times.corec.cong_base
(* is an equivalence: *)
thm times.corec.cong_refl times.corec.cong_sym times.corec.cong_trans
(* ... and is compatible with the friends: *)
thm times.corec.cong_SCons times.corec.cong_plus times.corec.cong_times
    

(* Concrete example proof: *)
lemma times_comm:  
"times xs1 xs2 = times xs2 xs1"
  apply (coinduction arbitrary: xs1 xs2 rule: stream.coinduct_upto)
  apply simp
  apply(subst plus_comm)
  apply(rule times.corec.cong_plus)
   apply(rule times.corec.cong_base) apply auto[]
   apply(rule times.corec.cong_base) apply auto[]
  done

(* More succinct: 
apply (coinduction arbitrary: xs1 xs2 rule: stream.coinduct_upto) 
  apply simp apply(subst plus_comm)
  apply(auto intro!: times.corec.cong_plus intro: times.corec.cong_base)
*)


end
  
