(* Programming and Reasoning with Infinite Data in Isabelle/HOL (POPL 2018) *)
(* Mathias Fleury, Andreas Lochbihler, Andrei Popescu *)

theory Corec_Internals imports Prelim
begin

(* To keep notation simple, I suppress polymorphism: *)  
codatatype stream = SCons (shd: nat) (stl: stream)

corec plus :: "stream \<Rightarrow> stream \<Rightarrow> stream" where 
  "plus xs ys = SCons (shd xs + shd ys) (plus (stl xs) (stl ys))"
    
friend_of_corec plus where 
  "plus xs ys = SCons (shd xs + shd ys) (plus (stl xs) (stl ys))"
using plus.code apply blast   
apply transfer_prover
done

(* Some notations: *)
type_synonym 'a \<Sigma> = "'a sig_stream_v0"
type_synonym 'a \<Sigma>s = "'a ssig_stream_v0"  
    
abbreviation Sig :: "nat \<times> 'a \<Rightarrow> 'a \<Sigma>" where "Sig \<equiv> stream.v0.Sig"
(* Marks the corecursive call: *)
abbreviation CALL :: "'a \<Rightarrow> 'a \<Sigma>s" where "CALL \<equiv> stream.v0.VLeaf" 
(* Marks the outer context, on top of the guard: *)
abbreviation OUT :: "'a \<Rightarrow> 'a \<Sigma>s" where "OUT \<equiv> stream.v0.VLeaf"
abbreviation CT :: "stream \<Rightarrow> 'a \<Sigma>s" where "CT \<equiv> stream.v0.CLeaf"
abbreviation Oper :: "'a \<Sigma>s \<Sigma> \<Rightarrow> 'a \<Sigma>s" where "Oper \<equiv> stream.v0.Oper"
abbreviation "GUARD \<equiv> Pair"
abbreviation CONS :: "nat \<Rightarrow> 'a \<Sigma>s \<Rightarrow> 'a \<Sigma>s"    
  where "CONS n x \<equiv> Oper (Sig (n,x))"

type_synonym 'a zoDest = "'a \<times> (nat \<times> 'a)"  

  
(* User spec: 
corec plus :: "stream \<Rightarrow> stream \<Rightarrow> stream" where 
  "plus xs ys = Cons (hd xs + hd ys) 
                     (plus (tl xs) (tl ys))"
*)  
(* Synthesize blueprint: *) 
term "(\<lambda>(xs, ys). OUT (GUARD (shd xs + shd ys) 
                             (CALL (stl xs, stl ys)))) 
::
  stream \<times> stream \<Rightarrow> (nat \<times> (stream \<times> stream)\<Sigma>s) \<Sigma>s"
(* Synthesize surface: *) 
term "(\<lambda>(xs, ys). OUT (GUARD (fst (snd xs) + fst (snd ys))
                             (CALL (snd (snd xs), snd (snd ys))))) 
:: 
  'a zoDest \<times> 'a zoDest \<Rightarrow> (nat \<times> ('a \<times> 'a)\<Sigma>s) \<Sigma>s"
  
fun expand :: "stream \<Rightarrow> stream zoDest" where 
  "expand xs = (xs,(shd xs, stl xs))"

  
(* It is proved that: 
-- (1) plus_blueprint = plus_surface \<circ> map_prod expand expand
-- (2) plus_surface is relationally parametric
*)

    


  
end